/**
 * Copyright (C) 2023-2024 Atmark Techno, Inc. All Rights Reserved.
 * SPDX-License-Identifier: Apache-2.0
 */

#include "command_line_utils.h"
#include "mqtt_handler.h"
#include "shadow_handler.h"
#include "shadow_factory.h"
#include "job_handler.h"
#include "agent_utils.h"
#include "agent_config.h"
#include "agent_log.h"

#define SHADOW_NAME "heart_beat"
#define HEARTBEAT_INTERVAL 300 // seconds
#define DEFAULT_ENDPOINT "a2cxtlwgru3aml-ats.iot.us-east-1.amazonaws.com"
#define DEFAULT_RECONNECT_INTERVAL 10 // seconds
#define MAX_RECONNECT_INTERVAL 60 * 60 * 24 // 1day (seconds)
#define LOGGING_HEARTBEAT_PER_DAY 4
#define LOGGING_HEARTBEAT_INTERVAL_COUNT static_cast<int>((24 * 60 * 60 / HEARTBEAT_INTERVAL) / LOGGING_HEARTBEAT_PER_DAY)

using namespace Aws::Crt;

int main(int argc, char *argv[])
{
    if (!checkAndCreatePidFile())
        exit(-1);
    ApiHandle apiHandle;

    AgentConfig agentConfig;
    AGENT_LOG_SET_CONSOLE_LOGLEVEL(agentConfig.get("output-console-loglevel"));
    AGENT_LOG_SET_LOGFILE_LOGLEVEL(agentConfig.get("output-logfile-loglevel"));

    String endpoint = agentConfig.get("endpoint");
    if (endpoint.empty())
    {
        AGENT_LOG_DEBUG("failed to get endpoint address from config file, use default endpoint");
        endpoint = DEFAULT_ENDPOINT;
    }
    bool isCommandExecAllowed = false;
    if (agentConfig.get("allow-exec-command") == "yes")
    {
        isCommandExecAllowed = true;
    }
    AGENT_LOG_DEBUG("exec-command feature status: %s",
                    isCommandExecAllowed ? "Enabled" : "Disabled");

    // command line arguments handling
    Utils::CommandLineUtils cmdUtils = Utils::CommandLineUtils(endpoint);
    cmdUtils.InitCommands();
    const char **const_argv = (const char **)argv;
    cmdUtils.SendArguments(const_argv, const_argv + argc);
    cmdUtils.StartLoggingBasedOnCommand(&apiHandle);

    // create instances
    String thingName = cmdUtils.GetCommandRequired("thing_name");
    MqttHandler* mqttHandler = nullptr;
    ShadowHandler* shadowHandler = new ShadowHandler(thingName, SHADOW_NAME);
    ShadowFactory* shadow_factory = new ShadowFactory(endpoint);
    JobHandler* jobHandler = nullptr;

    // MQTT setup
    std::shared_ptr<Mqtt::MqttConnection> connection = cmdUtils.BuildMQTTConnection();

    mqttHandler = new MqttHandler(thingName, connection);
    if (mqttHandler == nullptr)
    {
        AGENT_LOG_ERROR("failed to create instance of MqttHandler");
        exit(-1);
    }

    int reconnectInterval = DEFAULT_RECONNECT_INTERVAL;

    while (true)
    {
        int errCode = mqttHandler->Connect();
        if (errCode == AWS_ERROR_SUCCESS)
        {
            break;
        }

        AGENT_LOG_WARN("failed to connect to AWS. reconnect after %d seconds", reconnectInterval);
        std::this_thread::sleep_for(std::chrono::seconds(reconnectInterval));

        /* If an unprovisioned device attempts to connect to Twin and fails,
         * extend the reconnection interval (up to 1 day)
         *
         * from https://docs.aws.amazon.com/iot/latest/developerguide/device-advisor-troubleshooting.html
         * The AWS_ERROR_MQTT_UNEXPECTED_HANGUP error can occur under
         * a variety of conditions, but is most often described as
         * being due to AWS thing role.
         */
        if (errCode == AWS_ERROR_MQTT_UNEXPECTED_HANGUP &&
            reconnectInterval <= MAX_RECONNECT_INTERVAL)
        {
            reconnectInterval = std::min(reconnectInterval * 2, MAX_RECONNECT_INTERVAL);
        }
    }

    // subscribe topics of IoT shadow
    if (!shadowHandler->Subscribe(connection))
    {
        AGENT_LOG_ERROR("failed to subscribe to IoT shadow");
        exit(-1);
    }

    // subscribe topics of IoT Jobs
    jobHandler = new JobHandler(thingName, isCommandExecAllowed, connection);
    if (jobHandler == nullptr)
    {
        AGENT_LOG_ERROR("failed to create instance of JobHandler");
        exit(-1);
    }
    if (!jobHandler->Subscribe())
    {
        AGENT_LOG_ERROR("failed to IoT jobs subscribe");
        exit(-1);
    }
    jobHandler->InitialJobExec();

    // get /etc/sw-versions before starting job thread
    String swVersions = readFileContents("/etc/sw-versions");

    // start job thread
    std::thread jobMainLoop(&JobHandler::MainLoop, jobHandler);

    // upload device info
    Shadow *device_info_shadow = shadow_factory->CreateShadow(ShadowPurpose::DeviceInfo, isCommandExecAllowed);
    if (!shadowHandler->UpdateShadow(device_info_shadow->get()))
    {
        AGENT_LOG_ERROR("failed to update shadow");
        exit(-1);
    }

    // main loop
    int times_sent_heartbeat = 0;
    Shadow *heart_beat_shadow = shadow_factory->CreateShadow(ShadowPurpose::HeartBeat);
    while (true)
    {
        // Resend device_info when /etc/sw-versions is updated without reboot
        // by SWUpdate with POST_ACTION=container
        String newSwVersions = readFileContents("/etc/sw-versions");
        if (newSwVersions != swVersions)
        {
            shadowHandler->UpdateShadow(device_info_shadow->get());
            swVersions = newSwVersions;
        }

        // upload heartbeat
        JsonObject heart_beat = heart_beat_shadow->get();
        if (!shadowHandler->UpdateShadow(heart_beat))
        {
            AGENT_LOG_ERROR("failed to update shadow");
            exit(-1);
        }
        if (times_sent_heartbeat == 0)
            AGENT_LOG_INFO("%s", heart_beat.View().WriteCompact().c_str());
        times_sent_heartbeat = (times_sent_heartbeat + 1) % LOGGING_HEARTBEAT_INTERVAL_COUNT;

        std::this_thread::sleep_for(std::chrono::seconds(HEARTBEAT_INTERVAL));
    }
    delete device_info_shadow;
    delete heart_beat_shadow;

    // disconnect
    AGENT_LOG_DEBUG("Disconnecting...");
    mqttHandler->Disconnect();
    AGENT_LOG_DEBUG("Disconnected!");

    delete shadow_factory;
    delete shadowHandler;
    delete mqttHandler;
    delete jobHandler;
    return 0;
}
